from __future__ import print_function
from builtins import str
import json
from ws4py.client.threadedclient import WebSocketClient
from .base import _PuSuBaseClient
from .errors import PyPuSuConnectionError

# Auto-patch ws4py with wsaccel if that's available
try:
    import wsaccel

    wsaccel.patch_ws4py()
except ImportError:
    pass

DEBUG = False


class _ThreadedWebSocketPassthrough(WebSocketClient):
    """
    Simple passthrough class to get messages from the ws4py threadedclient.
    """

    def __init__(self, parent, url):
        super(_ThreadedWebSocketPassthrough, self).__init__(
            url,
            protocols=['http-only']
        )
        self.parent = parent

    def closed(self, code, reason=None):
        """
        Called when connection is closed

        :param int code: Error code
        :param str reason: Human readable message
        """

        self.parent._on_close(code, reason)

    def received_message(self, message):
        """
        Called when server sends us a message

        :param ws4py.TextMessage message: The message data
        """

        self.parent._received_message(message)


class PuSuClient(_PuSuBaseClient):
    """
    PuSu Engine client using the ws4py builtin client, not terribly impressive,
    but doesn't depend on anything else.

    .. code-block:: python
        from pypusy.threaded import PuSuClient

        client = PuSuClient("ws://127.0.0.1:55000")

        def listener(msg):
            # Gotcha: will be called in a different thread.
            print(msg)

        client.subscribe("my-channel", listener)

        # If you want to receive your own messages (which normally you don't)
        # you could sleep and wait for the subscription to go through.

        client.publish("some-channel", "data")
    """

    def __init__(self, *args):
        super(PuSuClient, self).__init__(*args)
        if DEBUG:
            self.debug = True

    def connect(self):
        """
        Connect to the WSPS server, starts the thread. Does not promise that
        connection is up immediately when this function returns.
        """

        if not self._client:
            self._client = _ThreadedWebSocketPassthrough(self, self.server)
        self._waiting_for = "hello"
        self._client.connect()
        self._wait()

    def close(self, code=1000, reason="", wait=True):
        """
        Disconnect from the server.

        :param int code: Error code
        :param str reason: Human readable message
        :param bool wait: If we should wait for the thread to stop
        """

        self._client.close(code, reason)

        if wait:
            self._client.run_forever()

    def _send(self, data):
        if self.debug:
            print("-> {}".format(data))
        try:
            self._client.send(data)
        except AttributeError:
            # ws4py throws AttributeErrors when the server disconnects ..
            raise PyPuSuConnectionError("Looks like we're disconnected")

    def _received_message(self, data):
        if self.debug:
            print("<- {}".format(data))
        self._on_receive(json.loads(str(data)))



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import json
from builtins import str
from queue import Queue, Empty

from .threaded import PuSuClient as ThreadedPuSuClient

DEBUG = False


class PuSuClient(ThreadedPuSuClient):
    """
    PuSu Engine client using the ws4py builtin client. Variant where reading
    messages takes polling, in case you want to avoid your code being called
    from another thread. Just make sure you poll often enough or the queue
    will use up your RAM.

    .. code-block:: python
        from pypusy.polling import PuSuClient

        client = PuSuClient("ws://127.0.0.1:55000")

        def listener(msg):
            print(msg)

        client.subscribe("my-channel", listener)
        client.publish("some-channel", "data")

        # Will call any subscribers if there were messages to deliver
        client.poll()
    """

    def __init__(self, *args):
        super(PuSuClient, self).__init__(*args)
        self._incoming_messages = Queue()
        if DEBUG:
            self.debug = True

    def poll(self):
        """
        Check for any new messages for us
        """
        count = 0
        while True:
            try:
                item = self._incoming_messages.get_nowait()
                count += 1
                self._on_receive(item)
            except Empty:
                break
        if self.debug:
            print("<< {} messages delivered from queue".format(count))

    def _received_message(self, data):
        if self.debug:
            print("<- {}".format(data))

        msg = json.loads(str(data))
        # Only publish messages should be paused, the rest is required for
        # normal function
        if msg["type"] == PuSuClient.TYPE_PUBLISH:
            self._incoming_messages.put(msg)
        else:
            self._on_receive(msg)



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import json
from builtins import object
from queue import Queue, Empty

DEBUG = False


class PuSuException(Exception):
    pass


class TimeoutExceeded(PuSuException):
    """
    Exception raised if the timeout was exceeded waiting for confirmation on
    authorize or subscribe.
    """
    pass


class Message(object):
    """
    A message sent via PuSu to subscribers of a channel.
    """

    def __init__(self, data=None):
        self.type = None
        self.channel = None
        self.content = None

        if data:
            for key in data:
                setattr(self, key, data[key])


class _PuSuBaseClient(object):
    """
    Shared functionality of all PuSu clients, defines the common API.
    """

    TYPE_HELLO = "hello"
    TYPE_AUTHORIZE = "authorize"
    TYPE_AUTHORIZATION_OK = "authorization_ok"
    TYPE_PUBLISH = "publish"
    TYPE_SUBSCRIBE = "subscribe"
    TYPE_SUBSCRIBE_OK = "subscribe_ok"

    def __init__(self, server=None, onclose=None):
        """
        :param server: Full URI of the destination PuSu server, e.g.
                       ws://localhost:52525/
        :param function onclose: Optional callback to run when connection is
                                 closed for any reason.
        """
        self.debug = DEBUG
        self.server = server
        self.onclose = onclose
        self.timeout = 5.0  # seconds

        self._wait_queue = Queue()
        self._waiting_for = False
        self._client = None
        self._subscriptions = {}

        if server:
            self.connect()

    def connect(self):
        raise NotImplementedError("{cls} does not implement connect()".format(
            cls=self.__class__.__name__
        ))

    def close(self):
        raise NotImplementedError("{cls} does not implement connect()".format(
            cls=self.__class__.__name__
        ))

    def _send(self, data):
        raise NotImplementedError(
            "{cls} does not implement _send_message()".format(
                cls=self.__class__.__name__
            ))

    def publish(self, channel, content):
        """
        Publish a message to a channel.

        :param str channel: Name of the channel to publish to
        :param str|int|float|dict|list content: Data to send
        """
        self._send(json.dumps({
            "type": self.TYPE_PUBLISH,
            "channel": channel,
            "content": content
        }))

    def subscribe(self, channel, callback):
        """
        Register a callback to messages from a channel.

        :param str channel: Name of the channel to subscribe to
        :param function callback: Function that processes messages from this
                                  channel
        """
        if channel not in self._subscriptions:
            self._waiting_for = self.TYPE_SUBSCRIBE_OK
            self._send(json.dumps({
                "type": self.TYPE_SUBSCRIBE,
                "channel": channel
            }))
            self._wait()
        self._subscriptions[channel] = callback

    def authorize(self, authorization):
        """
        Process your authorization.

        :param str authorization: The authorization to present to the server.
        """
        self._waiting_for = self.TYPE_AUTHORIZATION_OK
        self._send(json.dumps({
            "type": self.TYPE_AUTHORIZE,
            "authorization": authorization
        }))
        self._wait()

    def _wait(self):
        try:
            if self.debug:
                print("Waiting for {}".format(self._waiting_for))
            self._wait_queue.get(timeout=self.timeout)
            if self.debug:
                print("Got {}".format(self._waiting_for))
        except Empty:
            raise TimeoutExceeded()
        finally:
            self._waiting_for = None

    def _on_receive(self, msg):
        if msg["type"] == self._waiting_for:
            self._wait_queue.put(True)

        if msg["type"] == self.TYPE_PUBLISH:
            self._deliver_message(msg)

    def _on_close(self, code, reason=None):
        if self.onclose:
            self.onclose(code, reason)

    def _deliver_message(self, msg):
        if msg["channel"] in self._subscriptions:
            self._subscriptions[msg["channel"]](msg["content"])



from io import open
from os import path

from setuptools import setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='pypusu',
    packages=['pypusu'],  # this must be the same as the name above
    version='1.0.6',
    description='Python client for PuSuEngine',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author='Janne Enberg',
    author_email='janne.enberg@lietu.net',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*,<4",
    url='https://github.com/PuSuEngine/pypusu',
    download_url='https://github.com/PuSuEngine/pypusu/tarball/v1.0.6',
    keywords=['pubsub', 'publisher', 'subscriber', 'messaging'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "Programming Language :: Python",
        "Topic :: Communications :: Chat",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)
